from .utils import _b2i, _b2s
from .utils import _read_table_or_header_data, _read_atoms_from_mp4
from typing import List, Union


NOT_IMPLEMENTED_PARSERS = set( ('dref',))

_MAX_LIST_PRINT = 10
def _trunc_print(v) -> str:
  if type(v) == bytes:
    return str(v)
  if type(v) == str:
    return str(v)
  if type(v) == list:
    if len(v) > _MAX_LIST_PRINT:
      list_string = "[%i,...,%i]"%(v[0], v[-1])
    else:
      list_string = '[' + ",".join(map(str, v)) +']'

    return '(List with %i elements: %s)'%(len(v), list_string)
  
  return str(v)

class MP4Data:
  def __init__(self, data, offs):
    self._parsed = _read_table_or_header_data(data, offs)

  def get_parsed_data(self):
    return self._parsed

  def __repr__(self):
    s = ''
    s += "Leaf Atom (%s) %i bytes\n"%(self._parsed['atomtype'],self._parsed['atomsize'])
    for k, v in self._parsed.items():
      if k in ['atomtype','atomsize']:
        continue
      s += "  %s : %s \n"%(k,_trunc_print(v))
    return s

class NoData(MP4Data):
  def __init__(self, *args, **kwargs):
    super(NoData, self).__init__(*args, **kwargs)

  def __repr__(self):
    s  = '----> WARNING: The data parser for \'%s\' has not been implemented yet!\n       Parsed results may be WRONG!\n\n'%self._parsed['atomtype']
    s += super().__repr__()
    return s

# _table_or_header_class_map = {
#   'stts' : MP4Data,
#   'stsd' : MP4Data,
#   'stsc' : MP4Data,
#   'stsz' : MP4Data,
#   'tkhd' : MP4Data,
#   'mvhd' : MP4Data,
#   'stco' : MP4Data
# }
_table_and_header_classes = set(['stts','stsd','stsc','stsz','tkhd','mvhd','stco'])
def _header_or_data_object_factory(atomtype):
    if atomtype in _table_and_header_classes:
        return MP4Data
    else:
        return NoData


def _is_header_type(t):
  return t[-2:] == 'hd'

OTHER_LEAF_TYPES = set(['prfl','crgn', 'tkhd','crgn','kmat','elst' ,'stts','stss','stsd','stsz','stco','stsc'])
def _is_leaf_type(t):
  if _is_header_type(t):
    return True
  if t in OTHER_LEAF_TYPES:
    return True
  return False
  
class MP4Atom:
    def __init__(
            self, 
            data : bytes,
            offs : int = None,
            shift : int = None,
            atom_type : str = None,
            is_root = False,
            check_offsets_avail = False
        ):
        """
        dat:
          A bit stream corresponding to an MP4 file

        offs: 
          Offset from where to start reading for this atom

        shift (optional):
          how large is the atom in bytes (dat[offs:offs+shift])

        atom_type:
          the type of atom (redundant - maybe remove)
        
        is_root:
           if this is a root node

        check_offsets_avail:
           whether to check if the offsets are available in the 
           loaded bytes. The intended use mainly works on the header,
           so the `mdat` may not be loaded (context where this is 
           useful is for instance streaming mp4 where we get the header
           in the first few chunks). If this is true, when trying to 
           parse the `moov` field it will throw an error if not all of 
           it is available.

        """
        self.data = data
        self._check_offsets_avail = check_offsets_avail 
        if offs is None:
          offs = 0
        self.offs = offs

        if atom_type is None:
          atom_type = _b2s(self.data[self.offs+4:self.offs+8])
          if is_root:
            atom_type = '(root)'

        self.atom_type = atom_type
        self.is_root = is_root
          
        if shift is None:
          if not self.is_root:
            shift = _b2i(self.data[self.offs:self.offs+4])
          else:
            shift = len(data)
          
        self.shift = shift
        self.children = None
        self.header = None
        self.is_leaf = _is_leaf_type(self.atom_type)
        self.get_children()
    
    @staticmethod
    def init_from_chunk(bytes_or_path_string : Union[bytes,str], head_chunk_max_size = 50000, check_offsets_avail= False):
      """ Factory method - create a parser either for a file or a set of bytes in memory

      Note that "chunk" in the name of this function is not related to the technical 
      term used in the documentation of Mpeg-4 and quicktime.

      Params:
        bytes_or_path_string:
          Either a bytes object or a string pointing to a file.
          If this is a string, then the file is oppened in binary read mode
          and a set of bytes are read into memory for further processing.

        head_chunk_max_size:
          Because MP4 files may be large, read a limmited ammount from the header 
          (assuming that the interesting parts are in the start of the file which is 
          usually the case for streamed data).
      """
      if type(bytes_or_path_string) == str:
        with open(bytes_or_path_string,'rb') as f:
          data = f.read(head_chunk_max_size)
      else:
        data = bytes_or_path_string
        
      atom = MP4Atom(data, 
                    is_root = True,
                    check_offsets_avail = check_offsets_avail)
      return atom

    def _raw_data(self):
        self.data[self.offs:self.offs+self.shift]

    def get_children(self) -> List[tuple]:
      if not self.is_leaf:
        if self.children is None:
          self.children = _read_atoms_from_mp4(self.data, 
                                              offs_start = self.offs,
                                              end_shift=self.shift-8,
                                              check_offsets_avail = self._check_offsets_avail
                                            )
      else:
        self.children = []
      return self.children
    
    def __getitem__(self, key : str) -> Union['MP4Atom', List['MP4Atom']]: 
      _at = []
      for _atom_type, _o, _s in self.children:
        if _atom_type == key:
          if _is_leaf_type(_atom_type):
            _m = _header_or_data_object_factory(_atom_type)(self.data, offs = _o)
          else:
            if _atom_type not in NOT_IMPLEMENTED_PARSERS:
              _m = MP4Atom(self.data,
                          offs = _o+8,
                          shift = _s,
                          atom_type = _atom_type,
                          check_offsets_avail = self._check_offsets_avail)
            else:
              _m = NoData(self.data, offs = _o)
          _at.append(_m)
      if len(_at) == 1:
        _at = _at[0]
      return _at

    def __repr__(self) -> str:
      # s = 'type: %s \n  Children: \n    '%self.atom_type + str(self.children)
      s = '(Atom of type \'%s\' with %i children\n'%(str(self.atom_type), len(self.children))
      for c, o, l in self.children:
        s += '  %s : (offs: %i, size: %i)\n'%(c, o, l)
      return s[:-1] + ')'




# 
#   The following util functions implement partially the parsing of
# different mp4 atoms (Boxes). They were not extensively tested
# and their implementation was based on a combination of the following 
# resources: 
# 
#  1. Apple quicktime page (accessed Sept. 2022) here : https://developer.apple.com/library/archive/documentation/QuickTime/QTFF/QTFFChap2/qtff2.html#//apple_ref/doc/uid/TP40000939-CH204-SW1
#  2. Cimarronsystems document (accessed Sept. 2022) here: http://www.cimarronsystems.com/wp-content/uploads/2017/04/Elements-of-the-H.264-VideoAAC-Audio-MP4-Movie-v2_0.pdf
# 
#  
def _b2i(b):
    return int.from_bytes(b, byteorder='big',signed=False)
def _b2s(b):
    return str(b,encoding='utf-8')

def _read_atoms_from_mp4(
        data,
        offs_start = 0,
        end_shift  = None,
        check_offsets_avail = False,
    ):
    """
    reads metadata and data from the first chunk of an MP4 file.
    
    The data are organized in "atoms" which are pre-pended with a 
    name and a size (totalling in most cases 8 bytes - 4 bytes 
    size and 4 bytes name)

    """

    offs_list = []
    ci = offs_start
    n_atoms = 0
    if end_shift is None:
        end_shift = len(data) - offs_start
    
    if check_offsets_avail:
        assert((offs_start + end_shift) <=len(data) )

    while ci < (offs_start + end_shift):
        # if ci > len(data):
            # raise Exception("You tried to access an incorrect memory address while parsing MP4 data.")

        atom_size = data[ci:ci+4]
        atom_name = data[ci+4:ci+8] 

        offs_atom = _b2i(atom_size)
        offs_list.append( tuple((_b2s(atom_name), ci, offs_atom)) )

        ci += offs_atom
        if offs_atom == 0:
            ci += 1
        n_atoms += 1
    return offs_list


def _read_sample_to_chunk_table(dat):
    c = 0
    triplets = []
    while c < len(dat):
        first_chunk = dat[c:c+4]
        samples_per_chunk = dat[c+4:c+8]
        sample_desc_id = dat[c+8:c+12]
        triplets.append((_b2i(first_chunk),_b2i(samples_per_chunk), _b2i(sample_desc_id)))
        c+=12
    return triplets

def _parse_common_header_structure(dat_with_header) -> dict:
    """
    Reads the first 12 bits which have common meaning for several
    atom types and returns a dict for parsing.
    """
    res = {}
    res['atomsize'] = _b2i(dat_with_header[0:4])
    res['atomtype'] = _b2s(dat_with_header[4:8])
    res['version_byte'] = dat_with_header[8]
    res['flags_bytes'] = dat_with_header[8:12]
    return res


def _parse_tkhd(tkhd_dat, _ch_dict):
    _ch_dict['description'] = 'Header of the track Atom'
    _ch_dict['creat_utc_time'] = _b2i(tkhd_dat[12:16])
    _ch_dict['modif_utc_time']      = _b2i(tkhd_dat[16:20])
    _ch_dict['track_id']          = _b2i(tkhd_dat[20:24])
    _ch_dict['reserved']          =      tkhd_dat[24:28]
    _ch_dict['duration']          = _b2i(tkhd_dat[28:32])
    # (for res['duration'] : Duration of track in movie coord system
    #   equals to the sum of durations of all track edits
    #   and if there is no edit list it is the sum of 
    #   sample durations.
    return _ch_dict

def _parse_mvhd(mvhd_dat, _ch_dict):
    _ch_dict['description']      = 'Header of the movie Atom'
    _ch_dict['creat_utc_time']   = _b2i(mvhd_dat[12:16])
    _ch_dict['modif_utc_time']   = _b2i(mvhd_dat[16:20])
    _ch_dict['timescale']        = _b2i(mvhd_dat[20:24])
    _ch_dict['duration']         = _b2i(mvhd_dat[24:28])
    _ch_dict['pref_rate']        = _b2i(mvhd_dat[28:32])
    _ch_dict['pref_volume']      = _b2i(mvhd_dat[32:34])
    _ch_dict['reserved']         =      mvhd_dat[34:44]
    _ch_dict['matrix']           =  [[_b2i(mvhd_dat[44 + (i + j*3)*4:44+(i+1+j*3)*4]) for i in range(0,3)] for j in range(0, 3)] 

    _remaining_fields = ['preview_time',
                       'preview_duration',
                       'poster_time',
                       'selection_time',
                       'selection_duration',
                       'current_time',
                       'next_track_id']
    for i, field_name in enumerate(_remaining_fields):
        _ch_dict[field_name] = _b2i(mvhd_dat[80+i*4:80+(i+1)*4])
    
    return _ch_dict


def _parse_stsz(stsz_dat, _ch_dict):
    _ch_dict['description'] = 'Sample size Atom. Contains a table of sample sizes for the parent track.'
    _ch_dict['sample_size'] = _b2i(stsz_dat[12:16])
    _ch_dict['num_entries'] = _b2i(stsz_dat[16:20])
    sample_size_table  = []; 
    c = 0
    while c < _ch_dict['num_entries']:
        sample_size_table.append(_b2i(stsz_dat[20+c*4:20+(c+1)*4]))
        c += 1

    _ch_dict['sample_size_table'] = sample_size_table
    return _ch_dict


def _read_sample_description_stsd(sample_table):
    """
    According to the quicktime ref "Sample Description Atoms" (part of the stsd desc.)
    the array is parsed as follows.
    
    [4-size][4-fmt][6-reserved][2 bytes -  index]
    """
    # sample_table = tab['stsd']
    num_sample_descr = int.from_bytes(sample_table[0:4],byteorder = 'big',signed = False) # 32 bits sample desc size
    data_fmt = sample_table[4:8] # 32 bits format
    reserved = sample_table[8:14] # all zeros if all good
    data_reference_index = sample_table[14:16] # index to correspond the sample to  a data reference
    res = {
        'num_sample_desc' : num_sample_descr,
         'data_fmt' : data_fmt, 
         'data_reference_index' : _b2i(data_reference_index)
    }
    res['description'] = 'The sample description table. Keeps track of how many compressed frames are stored in each sample.'
    return res

def _read_sample_description_stts(time_to_sample_table):
    res = {
        'sample_duration' :  _b2i(time_to_sample_table[0:4]),
        'sample_count' :  time_to_sample_table[4:8]
    }
    return res

def _read_stco_table(table_dat):
    """
    A table containing the sample offsets. 
    It should be equal in number to the number
     of entries defined in the num_entries field.
    """
    r = []
    c = 0
    while c < len(table_dat):
        t = _b2i(table_dat[0+c:4+c])
        r.append(t)
        c+=4
    return r

special_parsers = {
    'stsz' : _parse_stsz,
    'tkhd' : _parse_tkhd,
    'mvhd' : _parse_mvhd
}

def _read_table_or_header_data(data, offset):
    """
    Tables and headers have special fields that need their own parsers
    """
    res = _parse_common_header_structure(data[offset:offset+12])
    _atomsize  = res['atomsize']
    _atomtype  = res['atomtype']

    if _atomtype in special_parsers: # re-reading some bits here
        table_data = data[offset:offset+_atomsize]
        res = special_parsers[_atomtype](table_data, res)
        return res

    # After "Flags: (12 bytes)"
    res['num_entries']  = _b2i(data[offset+12:offset+16])
    table_dat = data[offset+16:offset+_atomsize]

    if res['atomtype'] == 'stsd':
        res['description'] = 'Sample description Atom'
        res['table_data'] = _read_sample_description_stsd(table_dat)
    if res['atomtype'] == 'stts':
        res['table_data'] = _read_sample_description_stts(table_dat)
    if res['atomtype'] == 'stsc':
        res['table_columns'] = ['first_chunk','sample_per_chunk','sample_description_id']
        res['table_data'] = _read_sample_to_chunk_table(table_dat)
    if res['atomtype'] == 'stco':
        res['description'] = 'Chunk offset data table.'
        res['table_data'] = _read_stco_table(table_dat)
    return res



import setuptools
import re
import os


VERSION=os.environ['YAPY_MP4_VERSION']
def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + '/__init__.py').read())
    return result.group(1)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yapymp4", # Replace with your own username
    version=VERSION,
    author="Charilaos Mylonas",
    author_email="mylonas.charilaos@gmail.com",
    description="Implementation of a parser for MP4 in pure python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mylonasc/yapyMP4.git",
    packages=setuptools.find_packages(),
    package_data ={'':['assets/html_css/*css','assets/html_css/*js']},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
